# -*- coding: utf-8 -*-

import re


class EncodingOptions(object):

    def __init__(self, options=None):
        options = options or {}
        self._options = {}

        self._options.update(options)

    def supports_option(self, option):
        return hasattr(self, '_convert_%s' % option)\
            and callable(getattr(self, '_convert_%s' % option))

    def __str__(self):
        params = []
        for option, value in self._options.iteritems():
            if value:
                option_list = getattr(self, '_convert_%s' % option)(value)
                params.append('%s' % ' '.join(option_list))

        # codecs should go before the presets so that the files will be matched successfully
        # all other parameters go after so that we can override whatever is in the preset
        codecs = []
        presets = []
        other = []
        for p in params:
            if re.match('codec', p):
                codecs.append(p)
            elif re.match('\-.pre', p):
                presets.append(p)
            else:
                other.append(p)

        params = codecs + presets + other

        params_string = ' '.join(params)
        if self['aspect'] and self['resolution']:
            params_string += '%s' % self._convert_aspect(self._calculate_aspect())

        return params_string

    def as_parameters(self):
        params = []
        for option, value in self._options.iteritems():
            if value:
                params.append(getattr(self, '_convert_%s' % option)(value))

        # codecs should go before the presets so that the files will be matched successfully
        # all other parameters go after so that we can override whatever is in the preset
        codecs = []
        presets = []
        other = []
        for p in params:
            if re.match('codec', p[0]):
                codecs.append(p)
            elif re.match('\-.pre', p[0]):
                presets.append(p)
            else:
                other.append(p)

        params = codecs + presets + other

        if self['aspect'] and self['resolution']:
            params += self._convert_aspect(self._calculate_aspect())

        return [i for l in params for i in l]

    @property
    def width(self):
        try:
            return int(self['resolution'].split('x')[0])
        except IndexError:
            return

    @property
    def height(self):
        try:
            return int(self['resolution'].split('x')[1])
        except IndexError:
            return

    def _convert_aspect(self, value):
        return ['-aspect', value]

    def _calculate_aspect(self):
        return float(self.width) / float(self.height)

    def _convert_video_codec(self, value):
        return ['-vcodec', value]

    def _convert_frame_rate(self, value):
        return ['-r', value]

    def _convert_resolution(self, value):
        return ['-s', value]

    def _convert_video_bitrate(self, value):
        return ['-b:v', self.k_format(value)]

    def _convert_audio_codec(self, value):
        return ['-acodec', value]

    def _convert_audio_bitrate(self, value):
        return ['-b:a', self.k_format(value)]

    def _convert_audio_sample_rate(self, value):
        return ['-ar', value]

    def _convert_audio_channels(self, value):
        return ['-ac', value]

    def _convert_video_max_bitrate(self, value):
        return ['-maxrate', self.k_format(value)]

    def _convert_video_min_bitrate(self, value):
        return ['-minrate', self.k_format(value)]

    def _convert_buffer_size(self, value):
        return ['-bufsize', self.k_format(value)]

    def _convert_video_bitrate_tolerance(self, value):
        return ['-bt', self.k_format(value)]

    def _convert_threads(self, value):
        return ['-threads', value]

    def _convert_duration(self, value):
        return ['-t', value]

    def _convert_video_preset(self, value):
        return ['-vpre', value]

    def _convert_audio_preset(self, value):
        return ['-apre', value]

    def _convert_file_preset(self, value):
        return ['-fpre', value]

    def _convert_keyframe_interval(self, value):
        return ['-g', value]

    def _convert_seek_time(self, value):
        return ['-ss', value]

    def _convert_screenshot(self, value):
        if value:
            return ['-vframes', '1', '-f', 'image2']

        return []

    def _convert_x264_profile(self, value):
        return ['-vprofile', value]

    def _convert_x264_preset(self, value):
        return ['-preset', value]

    def _convert_watermark(self, value):
        return ['-i', value]

    def _convert_custom(self, value):
        return value

    def k_format(self, value):
        if re.match('^.*k$', value):
            return value

        return '%sk' % value

    def __getitem__(self, item):
        return self._options.get(item)

    def __setitem__(self, key, value):
        self._options[key] = value



# -*- coding: utf-8 -*-

import subprocess
import errno

VALID_BINARIES = ['ffmpeg', 'avconv']
BINARY = None

# Figuring out which binary to use
for binary in VALID_BINARIES:
    try:
        subprocess.check_output([binary], stderr=subprocess.STDOUT)
    except OSError as e:
        if e.errno == errno.ENOENT:
            continue

        BINARY = binary
        break
    except subprocess.CalledProcessError:
        BINARY = binary
        break

if BINARY is None:
    raise Exception('No valid binary could be found. '
                    'One of %s must be installed.'
                    % ', '.join(VALID_BINARIES))



# -*- coding: utf-8 -*-

import math
import re
import subprocess
import select
import fcntl
import os
import pexpect
from .encoding_options import EncodingOptions
from . import BINARY


class Transcoder(object):

    timeout = 30

    def __init__(self, movie, output_file,
                 options=EncodingOptions(), transcoder_options=None):
        """
        :param movie: a Movie instance
        :type movie: Movie
        :param options: an EncodinOptions instance
        :type options: EncodingOptions
        """
        transcoder_options = transcoder_options or {}

        self.movie = movie
        self.output_file = output_file

        if isinstance(options, (basestring, EncodingOptions)):
            self.raw_options = options
        elif isinstance(options, dict):
            self.raw_options = EncodingOptions(options)
        else:
            raise Exception('Unknown options format %s, '
                            'should be either EncodingOptions, str or dict'
                            % type(options))

        self.transcoder_options = transcoder_options
        self.errors = []

        self.apply_transcoder_options()

    def run(self):
        for progress in self._transcode_movie():
            yield progress

    def apply_transcoder_options(self):
        self.transcoder_options['validate'] = self.transcoder_options.get('validate', True)

        if self.movie.aspect_ratio is None:
            return

        preserve_aspect_ratio = self.transcoder_options.get('preserver_aspect_ratio')
        if preserve_aspect_ratio == 'width':
            new_height = self.raw_options['width'] / self.movie.aspect_ratio
            if new_height % 2 == 0:
                new_height = math.ceil(new_height)
            else:
                new_height = math.floor(new_height)

            if new_height % 2 == 1:
                new_height += 1

            self.raw_options['resolution'] = '%sx%s' % (self.raw_options['width'], new_height)
        elif preserve_aspect_ratio == 'height':
            new_width = self.raw_options['height'] * self.movie.aspect_ratio
            if new_width % 2 == 0:
                new_width = math.ceil(new_width)
            else:
                new_width = math.floor(new_width)

            if new_width % 2 == 1:
                new_width += 1

            self.raw_options['resolution'] = '%sx%s' % (new_width, self.raw_options['height'])

    def _transcode_movie(self):
        command = [BINARY, '-y', '-i', self.movie.path] + self.raw_options.as_parameters() + [self.output_file]

        thread = pexpect.spawn(subprocess.list2cmdline(command))
        cpl = thread.compile_pattern_list([
            pexpect.EOF,
            "time=(\d+):(\d+):(\d+.\d+)"
        ])

        while True:
            i = thread.expect_list(cpl, timeout=5)
            if i == 0:
                yield 0.0
                break
            elif i == 1:
                m = thread.match

                time = int(m.group(1)) * 3600 + int(m.group(2)) * 60 + float(m.group(3))

                progress = time / self.movie.duration

                yield progress

        thread.close()




# -*- coding: utf-8 -*-

import os
import re
import subprocess
import errno
import dateutil.parser as date_parser

from . import BINARY
from .encoding_options import EncodingOptions
from .transcoder import Transcoder


class Movie(object):

    def __init__(self, path):
        if not os.path.exists(path):
            raise OSError((errno.ENOENT, 'File %s does not exist' % path))

        output = self._execute_command(path)

        self.path = os.path.realpath(path)
        self.name = os.path.basename(path)
        self.size = os.path.getsize(path)
        self.container = re.search('Input #\d+,\s*(\S+),\s*from', output).group(1)
        self.time = 0.0
        self.creation_time = None
        self.bitrate = None
        self.rotation = None
        self.frame_rate = None
        self.video_stream = None
        self.audio_stream = None
        self.colorspace = None
        self.video_bitrate = None
        self.video_codec = None
        self.sar = None
        self.dar = None
        self.audio_bitrate = None
        self._audio_channels = None
        self.audio_codec = None
        self.audio_sample_rate = None
        self.invalid = False

        m = re.search('Duration: (\d{2}):(\d{2}):(\d{2}\.\d{2})', output)
        if m:
            self.duration = int(m.group(1)) * 60 * 60 + int(m.group(2)) * 60 + float(m.group(3))

        m = re.search('start: (\d*\.\d*)', output)
        if m:
            self.time = float(m.group(1))

        m = re.search('creation_time +: +(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})', output)
        if m:
            self.creation_time = date_parser.parse(m.group(1))

        m = re.search('bitrate: (\d*)', output)
        if m:
            self.bitrate = int(m.group(1))

        m = re.search('rotate {1,}: {1,}(\d*)', output)
        if m:
            self.rotation = int(m.group(1))

        matches = re.findall('Video: (.*)', output)
        if matches:
            if len(matches) == 1:
                self.video_stream = matches[0]
            else:
                # Multiple video streams found, figuring out the right one
                for m in matches:
                    video_codec = re.match('((?:\([^()]*\)|[^,])+)', m).group(1)
                    # The first video stream in some mp4 files
                    # (like iTunes one) have a png file as
                    # first video stream, so we ignore it
                    if video_codec == 'png':
                        continue

                    self.video_stream = m
                    break

        m = re.search('Audio: (.*)', output)
        if m:
            self.audio_stream = m.group(1)

        if self.video_stream:
            # regexp to handle "yuv420p(tv, bt709)" colorspace etc from http://goo.gl/6oi645
            commas_except_in_parenthesis = '(?:\([^()]*\)|[^,])+'
            video_stream_info = re.findall(commas_except_in_parenthesis, self.video_stream)
            self.video_codec = video_stream_info[0].strip()
            self.colorspace = video_stream_info[1].strip()
            resolution = video_stream_info[2].strip()
            video_bitrate = video_stream_info[3].strip()

            m = re.match('\A(\d+) kb/s\Z', video_bitrate)
            if m:
                self.video_bitrate = int(m.group(1))

            # Trying to figure out the frame rate
            for info in video_stream_info[4:]:
                m = re.search('(\d*\.?\d*)\s?(fps|tbr)', info)
                if m:
                    self.frame_rate = float(m.group(1))
                    break

            self.resolution = resolution.split(' ')[0]

            m = re.match('SAR (\d+:\d+)', self.video_stream)
            if m:
                self.sar = m.group(1)

            m = re.match('DAR (\d+:\d+)', self.video_stream)
            if m:
                self.sar = m.group(1)

        if self.audio_stream:
            audio_stream_info = re.split('\s?,\s?', self.audio_stream)
            self.audio_codec = audio_stream_info[0]
            audio_sample_rate = audio_stream_info[1]
            self._audio_channels = audio_stream_info[2]
            audio_bitrate = audio_stream_info[4]

            m = re.match('\A(\d+) kb/s\Z', audio_bitrate)
            if m:
                self.audio_bitrate = int(m.group(1))

            self.audio_sample_rate = int(re.match('(\d*)', audio_sample_rate).group(1))

        if self.video_stream is None and self.audio_stream is None\
                or 'is not supported' in output\
                or 'could not find codec parameters' in output:
            self.invalid = True

    def transcode(self, output_file, options=EncodingOptions(), transcoder_options=None):
        transcoder = Transcoder(self, output_file, options, transcoder_options)

        for progress in transcoder.run():
            yield progress

    @property
    def audio_channels(self):
        if not self._audio_channels:
            return None
        elif 'channels' in self._audio_channels:
            return int(re.match('(\d*)', self._audio_channels).group(1))
        elif 'mono' in self._audio_channels:
            return 1
        elif 'stereo' in self._audio_channels:
            return 2
        elif '5.1' in self._audio_channels:
            return 6

    @property
    def width(self):
        try:
            return int(self.resolution.split('x')[0])
        except (NameError, IndexError):
            return

    @property
    def height(self):
        try:
            return int(self.resolution.split('x')[1])
        except (NameError, IndexError):
            return

    @property
    def aspect_ratio(self):
        return self._aspect_from_dar() or self._aspect_from_dimensions()

    @property
    def pixel_aspect_ratio(self):
        return self._aspect_from_sar() or 1

    def is_valid(self):
        return not self.invalid

    def _aspect_from_dar(self):
        if not self.dar:
            return

        w, h = self.dar.split(':')
        aspect = float(w) / float(h)

        if not aspect:
            return

        return aspect

    def _aspect_from_sar(self):
        if not self.sar:
            return

        w, h = self.sar.split(':')
        aspect = float(w) / float(h)

        if not aspect:
            return

        return aspect

    def _aspect_from_dimensions(self):
        aspect = float(self.width) / float(self.height)

        if not aspect:
            return

        return aspect

    def _execute_command(self, path):
        command = [BINARY, '-i', path]

        try:
            subprocess.check_output(command, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            return e.output



# -*- coding: utf-8 -*-

from unittest import TestCase
from lampwick.encoding_options import EncodingOptions


class TestMovie(TestCase):

    def test_init(self):
        base_options = {
            'video_codec': 'h264'
        }
        options = EncodingOptions(base_options)

        print(options)
        print(options.as_parameters())



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

__version__ = '0.1.1'

setup(
    name='lampwick',
    license='MIT',
    version=__version__,
    description='Simple wrappper around ffmpeg commands.',
    author='Sébastien Eustace',
    author_email='sebastien.eustace@gmail.com',
    url='https://github.com/sdispater/lampwick',
    download_url='https://github.com/sdispater/lampwick/archive/%s.tar.gz' % __version__,
    packages=find_packages(),
    install_requires=['nose', 'pexpect', 'dateutils'],
    tests_require=['nose'],
    test_suite='nose.collector',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
